/*
 * Copyright (c) 2008, 2011 The_xmloperator_project.  All rights reserved.
 * See terms of use on the project's site (http://www.xmloperator.net/reactive/).
 */
package org.xmloperator.reactive;

import java.util.List;

import org.xmloperator.reactive.exception.DisconnectedEntityException;
import org.xmloperator.reactive.exception.EntityAlreadyConnectedException;
import org.xmloperator.reactive.exception.NotMessageEventException;
import org.xmloperator.reactive.exception.NotOwnEventException;
import org.xmloperator.reactive.exception.NotOwnThreadException;
import org.xmloperator.reactive.exception.ThreadAlreadyAttachedException;

/**
 * A ReactiveEntity is attached to an Agent and its thread.
 * The Agent communicates with a ReactiveCommunity through this interface.
 * Rule : an Agent must not publish the reference in memory of any ReactiveEntity.
 */
public interface ReactiveEntity {

	/**
	 * Status of a ReactiveEntity that is not connected.
	 * Status can change to STATUS_ACTIVE.
	 */
	public final static int STATUS_NOT_CONNECTED = 0;
	
	/**
	 * Status of a ReactiveEntity that is connected, active and running.
	 * Status can change to STATUS_SUSPENDED, STATUS_STOPPED, STATUS_WAITING
	 *	or STATUS_NOT_CONNECTED.
	 */
	public final static int STATUS_ACTIVE = 1;
	
	/**
	 * Status of a ReactiveEntity that is connected and suspended.
	 * Status can change to STATUS_ACTIVE.
	 */
	public final static int STATUS_SUSPENDED = 2;
	
	/**
	 * Status of a ReactiveEntity that is connected and stopped.
	 * Status can change to STATUS_ACTIVE.
	 */
	public final static int STATUS_STOPPED = 3;
	
	/**
	 * Status of a ReactiveEntity that is connected, active but waiting.
	 * This is the case for launching an agent.
	 * Status can change to STATUS_ACTIVE.
	 */
	public final static int STATUS_WAITING = 4;
	
	/**
	 * Attach a Thread to this entity.
	 * @param thread a Thread to attach to this entity
	 * @exception ThreadAlreadyAttachedException is thrown if a Thread is already
	 * 	attached to this ReactiveEntity.
	 */
	public void setThread(Thread thread) throws ThreadAlreadyAttachedException;
	
	/**
	 * Returns the entityId associated to this ReactiveEntity.
	 * The entityId is internalized and unique within the virtual machine.
	 * @return the entityId associated to this ReactiveEntity.
	 */
	public String getEntityId();
	
	/**
	 * Returns the current status of this ReactiveEntity.
	 * @return the current status of this ReactiveEntity.
	 */
	public int getStatus();
	
	/**
	 * Set the entity status to STATUS_WAITING.
	 */
	public void setStatusWaiting();
	
	/**
	 * Set the entity status to STATUS_ACTIVE.
	 */
	public void setStatusActive();
	
	/**
	 * @return the ReactiveCommunity on which it is connected,
	 * 	null if it is not connected.
	 */	
	public ReactiveCommunity getCommunity();

	/**
	 * Connects itself to the given community. This entity must not be be
	 * already connected. The connection occurs during the inactive part of an
	 * instant. This method can only be called by its owner thread.
	 * 
	 * @param community
	 *            a ReactiveCommunity to which to connect.
	 * @param withExclusiveEntityGroup
	 *            if true then an exclusive EntityGroup and an exclusive message
	 *            Event are created or recovered.
	 * @exception NotOwnThreadException
	 *                if the calling Thread is not the attached one.
	 * @throws EntityAlreadyConnectedException
	 *             if this entity is already connected.
	 */
	public void connect(ReactiveCommunity community, boolean withExclusiveEntityGroup)
			throws NotOwnThreadException, EntityAlreadyConnectedException;
	
	/**
	 * @return the exclusive message Event of this entity if any, null elsewhere.
	 */
	public Event getExclusiveMessageEvent();
	
	/**
	 * Generates the given Event within the current instant.
	 * Precondition : The ReactiveEntity must be connected.
	 * This method can only be called by its owner thread.
	 * @param event an Event to generate.
	 * @exception DisconnectedEntityException if this ReactiveEntity is
	 *   not connected.
	 * @exception NotOwnThreadException if the calling Thread is not
	 *   the attached one.
	 * @exception NotOwnEventException if the event doesn't belong to the community.
	 */
	public void generateEvent(Event event)
		throws
			DisconnectedEntityException,
			NotOwnThreadException,
			NotOwnEventException;

	/**
	 * Generates a message Event for broadcasting at the beginning of the next
	 * instant. Precondition : The ReactiveEntity must be connected. This method
	 * can only be called by its owner thread.
	 * 
	 * @param messageEvent
	 *            a message Event to broadcast.
	 * @param message
	 *            a message. It may be null.
	 * @exception DisconnectedEntityException
	 *                if this ReactiveEntity is not connected.
	 * @exception NotOwnThreadException
	 *                if the calling Thread is not the attached one.
	 * @exception NotOwnEventException
	 *                if the Event doesn't belong to the community.
	 * @throws NotMessageEventException if the Event is not a message Event.
	 */
	public void generateMessage(Event messageEvent, Message message)
		throws
			DisconnectedEntityException,
			NotOwnThreadException,
			NotOwnEventException,
			NotMessageEventException;

	/**
	 * Generates a message Event with an ActionMessage for being processed during the next
	 *   instant transition.
	 * Precondition : The ReactiveEntity must be connected.
	 * This method can only be called by its owner thread.
	 * @param action : an Action.
	 * @return the generated ActionMessage.
	 * @exception DisconnectedEntityException if this ReactiveEntity is
	 *   not connected.
	 * @exception NotOwnThreadException if the calling Thread is not
	 *   the attached one.
	 */
	public ActionMessage generateActionMessage(Action action)
		throws
			DisconnectedEntityException,
			NotOwnThreadException;

	/**
	 * Generates an Action for being executed during the next instant
	 * transition. Precondition : The ReactiveEntity must be connected. This
	 * method can only be called by its owner thread.
	 * 
	 * @param action
	 *            : an Action.
	 * @exception DisconnectedEntityException
	 *                if this ReactiveEntity is not connected.
	 * @exception NotOwnThreadException
	 *                if the calling Thread is not the attached one.
	 */
	public void generateAction(Action action)
		throws
			DisconnectedEntityException,
			NotOwnThreadException;

	/**
	 * Suspend itself for the current instant.
	 * Precondition : The ReactiveEntity must be connected.
	 * It could yet react on events at the current instant.
	 * If some events (including message events) occur on the same instant then returns them.
	 * An Event that was generated by itself during the same instant may appear.
	 * 	This occurs when this entity is the first one to generate this Event
	 * 	during this instant.
	 * If none then stops and returns an empty list.
	 * This method can only be called by its owner thread.
	 * @return some occurring Events, if any.
	 * @exception DisconnectedEntityException if this ReactiveEntity is
	 *   not connected.
	 * @exception NotOwnThreadException if the calling Thread is not
	 *   the attached one.
	 */
	public List<Event> suspend()
		throws
			DisconnectedEntityException,
			NotOwnThreadException;
	
	/**
	 * Stops itself for the current instant.
	 * It will not react any more on the current instant.
	 * Precondition : The ReactiveEntity must be connected.
	 * Returns at the beginning of the next instant.
	 * This method can only be called by its owner thread.
	 * @exception DisconnectedEntityException if this ReactiveEntity is
	 *   not connected.
	 * @exception NotOwnThreadException if the calling Thread is not
	 *   the attached one.
	 */
	public void stop()
		throws
			DisconnectedEntityException,
			NotOwnThreadException;
	
	/**
	 * Disconnect itself from its current ReactiveCommunity.
	 * This method can only be called by its owner thread.
	 * @exception NotOwnThreadException if the calling Thread is not
	 *   the attached one.
	 */
	public void disconnect()
		throws NotOwnThreadException;
}
