/*
 * Copyright (c) 2008, 2011 The_xmloperator_project. All rights reserved.
 * See terms of use on the project's site (http://www.xmloperator.net/reactive/).
 */
package org.xmloperator.reactive.demo.land.agent;

import java.util.ArrayList;
import java.util.List;

import org.xmloperator.reactive.ActionMessage;
import org.xmloperator.reactive.Event;
import org.xmloperator.reactive.LaunchingContext;
import org.xmloperator.reactive.Message;
import org.xmloperator.reactive.ReactiveCommunity;
import org.xmloperator.reactive.agent.ReactiveAgentImpl;

import org.xmloperator.reactive.demo.land.AgentBody;
import org.xmloperator.reactive.demo.land.AgentBodyMessage;
import org.xmloperator.reactive.demo.land.AgentEnergy;
import org.xmloperator.reactive.demo.land.AgentFamily;
import org.xmloperator.reactive.demo.land.AgentLocation;
import org.xmloperator.reactive.demo.land.LandBoard;
import org.xmloperator.reactive.demo.land.Move;
import org.xmloperator.reactive.demo.land.NavigationStrategy;
import org.xmloperator.reactive.demo.land.Station;
import org.xmloperator.reactive.demo.land.Stations;
import org.xmloperator.reactive.demo.land.VisitMemory;
import org.xmloperator.reactive.demo.land.action.AgentEnergyAction;
import org.xmloperator.reactive.demo.land.action.AgentFamilyAction;
import org.xmloperator.reactive.demo.land.action.AgentMoveAction;
import org.xmloperator.reactive.demo.land.action.AgentQuitLocationAction;
import org.xmloperator.reactive.demo.land.impl.AgentBodyMessageImpl;
import org.xmloperator.reactive.demo.land.impl.VisitMemoryImpl;

import org.xmloperator.reactive.exception.IncorrectKeyException;
import org.xmloperator.reactive.exception.ReactiveException;

/**
 * An agent that goes from stations to stations, consumes energy and performs
 * cell division/fusion.
 */
public class BodyAgent extends ReactiveAgentImpl {
	
	/**
	 * AgentBody implementation.
	 */
	private class AgentBodyImpl implements AgentBody {
		
		///////////////////////////
		// AgentBody implementation
		///////////////////////////
		
		public String getEntityId() {
			return BodyAgent.this.entity.getEntityId();
		}

		public String getAgentId() {
			return BodyAgent.this.agentId;
		}
		
		public Event getExclusiveMessageEvent() {
			return BodyAgent.this.entity.getExclusiveMessageEvent();
		}

		public final AgentFamily getAgentFamily() {
			return BodyAgent.this.agentFamily;
		}
		
		public final int getCurrentEnergy(long familyKey) {
			if (familyKey != BodyAgent.this.agentFamilyKey)
				throw new IncorrectKeyException();
			return BodyAgent.this.currentEnergy;
		}
		
		////////////////////////
		// Object implementation
		////////////////////////
		
		public String toString() {
			return BodyAgent.this.agentId;
		}
	}

	/**
	 * AgentEnergy implementation.
	 */
	private class AgentEnergyImpl implements AgentEnergy {

		/////////////////////////////
		// AgentEnergy implementation
		/////////////////////////////
		
		public int getCurrentEnergy() {
			return BodyAgent.this.currentEnergy;
		}
		
		public void addEnergy(int energy) {
			BodyAgent.this.currentEnergy += energy;
		}

		////////////////////////
		// Object implementation
		////////////////////////
		
		public String toString() {
			return BodyAgent.this.agentFamily + "."
					+ BodyAgent.this.entity.getEntityId() + ".energy="
					+ BodyAgent.this.currentEnergy;
		}
	}

	/**
	 * AgentLocation implementation.
	 */
	private final class AgentLocationImpl implements AgentLocation {
		final Stations stations;
		final VisitMemory visitMemory;
		private int currentStationIndex = -1;
		private Station currentStation = null;

		AgentLocationImpl(Stations stations) {
			if (stations == null)
				throw new IllegalArgumentException();
			this.stations = stations;
			this.visitMemory = new VisitMemoryImpl(this.stations.getGraph()
					.getStationCount());
		}

		/**
		 * Set the location of the Agent.
		 * 
		 * @param newStationIndex
		 *            the index of the destination station or -1 if none.
		 * @return the Move if any, null elsewhere.
		 */
		private Move setStation(int newStationIndex) {
			int oldStationIndex = this.currentStationIndex;
			if (newStationIndex == oldStationIndex)
				return null;
			this.currentStationIndex = newStationIndex;
			this.currentStation = newStationIndex >= 0 ? this.stations
					.getStation(newStationIndex) : null;
			return new Move(oldStationIndex, newStationIndex);
		}

		///////////////////////////////
		// AgentLocation implementation
		///////////////////////////////

		public final AgentBody getAgentBody() {
			return BodyAgent.this.agentBody;
		}
		
		public final Station getCurrentStation() {
			return this.currentStation;
		}
		
		public final Station getFutureStation(NavigationStrategy navigationStrategy) {
			if (navigationStrategy == null)
				throw new IllegalArgumentException();
			if (this.currentStationIndex < 0)
				throw new IllegalStateException();
			int newStationIndex = navigationStrategy.nextStationIndex(
					this.currentStationIndex, this.visitMemory);
			if (newStationIndex == this.currentStationIndex)
				throw new ReactiveException("Not a navigation");
			return this.stations.getStation(newStationIndex);
		}
		
		public Move move(Station destination) {
			if (destination == null)
				throw new IllegalArgumentException();
			if (this.currentStation != null && !this.currentStation.isConnectedTo(destination))
				return null;
			int newStationIndex = destination.getStationIndex();
			if (this.currentStationIndex < 0)
				this.visitMemory.visit(newStationIndex);
			return this.setStation(newStationIndex);
		}

		public Move quit() {
			return (this.currentStationIndex >= 0) ? this.setStation(-1) : null;
		}
		
		public VisitMemory getVisitMemory() {
			return this.visitMemory;
		}
		
		////////////////////////
		// Object implementation
		////////////////////////
		
		public String toString() {
			return "S" + currentStationIndex;
		}
	}

	/*
	 * Maximum energy capital of this agent in resource units.
	 */
	public static final int maximumEnergy = 100;
	
	/*
	 * Maximum energy supply in an instant.
	 */
	public static final int maximumSupply = 20;
	
	/*
	 * Basic expense in an instant.
	 */
	public static final int basicExpense = 5;
	
	/*
	 * Moving expense in an instant. It is added to the basic expense.
	 */
	public static final int movingExpense = 5;
	
	private final String agentId;
	private final AgentFamily agentFamily;
	private final long agentFamilyKey;
	private final GenePool genePool;
	private final Station initialStation;
	private final StrategyStat strategyStat;
	private final BehaviorStat behaviorStat;
	private final long globalSeed;
	private final List<List<String>> roundTrace;
	private int currentEnergy;
	private final AgentBody agentBody = new AgentBodyImpl();;
	private final AgentEnergy agentEnergy = new AgentEnergyImpl();
	private final AgentBodyMessage leavingMessage = new AgentBodyMessageImpl(
			this.agentBody, AgentBodyMessage.leavingStation);
	private final AgentBodyMessage enteringMessage = new AgentBodyMessageImpl(
			this.agentBody, AgentBodyMessage.enteringStation);
	private final AgentLocation agentLocation;
	private final StringBuffer lifeTrace = null;//new StringBuffer();
	private int descendantCount = 0;
	private int divisionCount = 0;
	private int mergeCount = 0;
	private int giftCount = 0;
	private int moveCount = 0;
	private int supplyCount = 0;
	
	public BodyAgent(String agentId, AgentFamily agentFamily,
			long agentFamilyKey, GenePool genePool, int initialEnergy,
			Station initialStation, StrategyStat strategyStat,
			BehaviorStat behaviorStat, long globalSeed,
			List<List<String>> roundTrace) {
		super(true);
		if (agentId == null || agentFamily == null || genePool == null
				|| initialEnergy < 0 || initialStation == null
				|| strategyStat == null || behaviorStat == null)
			throw new IllegalArgumentException();
		this.agentId = agentId;
		this.agentFamily = agentFamily;
		this.agentFamilyKey = agentFamilyKey;
		this.genePool = genePool;
		this.currentEnergy = initialEnergy;
		this.initialStation = initialStation;
		this.strategyStat = strategyStat;
		this.behaviorStat = behaviorStat;
		this.globalSeed = globalSeed;
		this.roundTrace = roundTrace;
		this.agentLocation = new AgentLocationImpl(initialStation.getStations());
		this.addParameter(agentId);
	}

	/**
	 * @return an AgentBody candidate for merger with this one.
	 */
	private AgentBody candidateForMerger() {
		if (this.currentEnergy > maximumEnergy / 2)
			return null;
		final Station currentStation = this.agentLocation.getCurrentStation();
		int minimumEnergy = 1 + maximumEnergy / 2;
		AgentBody candidate = null;
		for (AgentBody agentBody : currentStation.getOccupantList())
			if (agentBody != this.agentBody
					&& agentBody.getAgentFamily() == this.agentFamily) {
				int energy = agentBody.getCurrentEnergy(this.agentFamilyKey);
				if (energy >= basicExpense + movingExpense
						&& energy > this.currentEnergy
						&& energy < minimumEnergy) {
					candidate = agentBody;
					minimumEnergy = energy;
				}
			}
		int candidateCount = 0;
		for (AgentBody agentBody : currentStation.getOccupantList())
			if (agentBody != this.agentBody
					&& agentBody.getAgentFamily() == this.agentFamily
					&& agentBody.getCurrentEnergy(this.agentFamilyKey) == minimumEnergy)
				candidateCount++;
		return candidateCount == 1 ? candidate : null;
	}
	
	/**
	 * Informs occupants of a Station of Agent arrival.
	 * @param station a Station.
	 */
	private final void notifyAgentArrival(Station station) {
		for (AgentBody occupant : station.getOccupantList())
			if (occupant != this.agentBody) {
				Event messageEvent = occupant.getExclusiveMessageEvent();
				if (messageEvent != null)
					this.entity.generateMessage(messageEvent, this.enteringMessage);
			}
	}
	
	/**
	 * Informs occupants of a Station of Agent departure.
	 * @param station a Station.
	 */
	private final void notifyAgentDeparture(Station station) {
		for (AgentBody occupant : station.getOccupantList())
			if (occupant != this.agentBody) {
				Event messageEvent = occupant.getExclusiveMessageEvent();
				if (messageEvent != null)
					this.entity.generateMessage(messageEvent, this.leavingMessage);
			}
	}
	
	///////////////////////////////
	// ReactiveAgent implementation
	///////////////////////////////
	
	public boolean reactToExclusiveMessage(Message message) {
		if (message instanceof AgentBodyMessage) {
			AgentBodyMessage agentBodyMessage = (AgentBodyMessage) message;
			Object information = agentBodyMessage.getInformation();
			if (information == AgentBodyMessage.enteringStation)
				;
			else if (information == AgentBodyMessage.leavingStation)
				;
			else if (information instanceof Integer) {
				int gift = ((Integer)information).intValue();
				if (gift > 0) {
					if (this.lifeTrace != null)
						this.lifeTrace.append("G");
					this.giftCount++;
					LandBoard landBoard = (LandBoard) this.entity
							.getCommunity().getBoard();
					this.entity.generateAction(new AgentEnergyAction(landBoard,
							this.agentEnergy, gift));
				}
			}
		}
		else if (message.isActionMessage())
			throw new ReactiveException(((ActionMessage)message).getException());
		return true;
	}

	public boolean newInstant() {
		final Station currentStation = this.agentLocation.getCurrentStation();
		if (currentStation == null)
			throw new IllegalStateException();
		/*
		 * Land end ?
		 */
		final LandBoard landBoard = (LandBoard) this.entity.getCommunity().getBoard();
		if (landBoard.isDone())
			return false;
		/*
		 * Round trace.
		 */
		final int instant = (int) this.entity.getCommunity().getInstant();
		synchronized (roundTrace) {
			while (roundTrace.size() <= instant) {
				List<String> instantTrace =  new ArrayList<String>();
				roundTrace.add(instantTrace);
			}
			List<String> instantTrace =  roundTrace.get(instant);
			String agentStatus = this.agentFamily + "." + this.agentId + " E"
					+ this.currentEnergy + " " + this.agentLocation;
			instantTrace.add(agentStatus);
		}
		/*
		 * Cell fusion.
		 */
		final AgentBody candidateForMerger = candidateForMerger();
		if (candidateForMerger != null) {
			if (this.lifeTrace != null)
				this.lifeTrace.append("F");
			this.mergeCount++;
			this.entity.generateMessage(
					candidateForMerger.getExclusiveMessageEvent(),
					new AgentBodyMessageImpl(
							this.agentBody,
							new Integer(this.currentEnergy)));
			return false;
		}
		/*
		 * Supply.
		 */
		final int currentStationOccupantCount = currentStation.getOccupantList().size();
		int supply = this.currentEnergy < maximumEnergy ? currentStation
				.getFlowSource() / currentStationOccupantCount : 0;
		if (supply > maximumSupply)
			supply = maximumSupply;
		if (this.lifeTrace != null)
			this.lifeTrace.append(supply == basicExpense ? "-"
					: supply > basicExpense ? "/" : "\\");
		this.supplyCount++;
		int totalSupply = supply;
		/*
		 * Cell division.
		 */
		final boolean isCellDivision = this.currentEnergy + totalSupply >= maximumEnergy;
		if (isCellDivision) {
			if (this.lifeTrace != null)
				this.lifeTrace.append("D");
			this.divisionCount++;
			String clonedAgentId = this.agentId + "." + ++descendantCount;
			int clonedAgentEnergy = (this.currentEnergy + totalSupply) / 4;
			totalSupply -= clonedAgentEnergy;
			long seed = 13 * this.globalSeed + 47 * instant;
			BodyAgent clonedAgent = new BodyAgent(clonedAgentId,
					this.agentFamily, this.agentFamilyKey,
					this.genePool.almostClone(seed), clonedAgentEnergy,
					currentStation, this.strategyStat, this.behaviorStat,
					this.globalSeed, this.roundTrace);
			this.entity.getCommunity().launchAgent(clonedAgent);
		}
		/*
		 * Basic expense.
		 */
		totalSupply -= basicExpense;
		if (this.currentEnergy + totalSupply < 0)
			return false;
		/*
		 * Decision to move
		 */
		if (isCellDivision
				|| (supply < basicExpense && this.currentEnergy + totalSupply >= movingExpense)
				|| (supply == basicExpense && this.currentEnergy + totalSupply >= maximumEnergy / 3)) {
			/*
			 * Moving expense.
			 */
			totalSupply -= movingExpense;
			/*
			 * Move
			 */
			if (this.lifeTrace != null)
				this.lifeTrace.append("M");
			this.moveCount++;
			final NavigationStrategy navigationStrategy = this.genePool
					.getNavigationStrategy();
			final Station newStation = this.agentLocation
					.getFutureStation(navigationStrategy);
			if (newStation == currentStation)
				throw new ReactiveException("Not a move");
			this.entity.generateAction(new AgentMoveAction(landBoard,
					this.agentLocation, newStation));
			this.notifyAgentDeparture(currentStation);
			this.notifyAgentArrival(newStation);
		}
		if (this.currentEnergy + totalSupply < 0)
			throw new ReactiveException("Negative energy");
		this.entity.generateAction(new AgentEnergyAction(landBoard,
				this.agentEnergy, totalSupply));
		return true;
	}

	///////////////////////
	// Agent implementation
	///////////////////////
	
	public void init(LaunchingContext launchingContext) {
		super.init(launchingContext);
		ReactiveCommunity community = this.entity.getCommunity();
		if (this.lifeTrace != null) {
			this.lifeTrace.append(this.agentBody);
			this.lifeTrace.append(" I");
			this.lifeTrace.append(community.getInstant());
			this.lifeTrace.append(" ");
			this.lifeTrace.append(this.initialStation);
			this.lifeTrace.append(" ");
		}
		LandBoard landBoard = (LandBoard) community.getBoard();
		this.entity.generateAction(new AgentFamilyAction(landBoard,
				this.agentBody, true));
		this.entity.generateAction(new AgentMoveAction(landBoard,
				this.agentLocation, this.initialStation));
		this.notifyAgentArrival(this.initialStation);
		this.entity.stop();
	}

	public void terminate() {
		final Station currentStation = this.agentLocation.getCurrentStation();
		ReactiveCommunity community = this.entity.getCommunity();
		LandBoard landBoard = (LandBoard) community.getBoard();
		if (this.trace != null) {
			if (this.lifeTrace != null) {
				this.lifeTrace.append(" I");
				this.lifeTrace.append(community.getInstant());
				this.lifeTrace.append(" ");
				this.lifeTrace.append("Q");
				this.trace.println(this.lifeTrace.toString());
			}
//			this.trace.println(this.agentLocation.getVisitMemory() + " "
//					+ this.agentBody.toString());
		}
		this.strategyStat.add(this.genePool);
		this.behaviorStat.add(this.divisionCount, this.mergeCount,
				this.giftCount, this.moveCount, this.supplyCount);
		this.entity.generateAction(new AgentQuitLocationAction(landBoard,
				this.agentLocation));
		this.entity.generateAction(new AgentFamilyAction(landBoard,
				this.agentBody, false));
		this.notifyAgentDeparture(currentStation);
		this.entity.stop();
		super.terminate();
	}
}
