package org.xmloperator.reactive.demo.land.agent;

import java.util.ArrayList;
import java.util.Collections;
import java.util.List;
import java.util.Random;

import org.xmloperator.reactive.ReactiveCommunity;
import org.xmloperator.reactive.agent.AgentImpl;
import org.xmloperator.reactive.agent.ReactiveFactoryImpl;

import org.xmloperator.reactive.demo.land.AgentFamily;
import org.xmloperator.reactive.demo.land.Graph;
import org.xmloperator.reactive.demo.land.LandBoard;
import org.xmloperator.reactive.demo.land.NavigationStrategy;
import org.xmloperator.reactive.demo.land.Station;
import org.xmloperator.reactive.demo.land.action.LandDoneAction;
import org.xmloperator.reactive.demo.land.impl.GraphImpl;
import org.xmloperator.reactive.demo.land.impl.LandBoardImpl;
import org.xmloperator.reactive.demo.land.impl.StationsImpl;

/**
 * Land demonstration.
 */
public class LandDemo extends AgentImpl {
	private final int stationCount = 16;
	private final int minPathCount = 3;
	private final int maxPathCount = 5;

	public LandDemo() {
		super(false);
	}

	/**
	 * A round consists in launching some agents belonging to some families, for
	 * some instants of life.
	 * 
	 * @param seed
	 *            a random number generator seed.
	 * @param stepCount
	 *            an instant count.
	 * @param roundTrace
	 *            a trace container. An instant is represented by a list of
	 *            agent status. An agent status is represented by a String.
	 */
	private void executeRound(long seed, int stepCount,
			List<List<String>> roundTrace) {
		final Random random = new Random(seed);
		final int[] graphLinks = org.xmloperator.reactive.demo.land.sphere.impl.StationsImpl
				.createUndirectedGraph(stationCount, minPathCount,
						maxPathCount, random);
		final Graph graph = new GraphImpl(stationCount, maxPathCount,
				graphLinks);
//		new DiscoveryNavigationStrategy(graph).test(random
//				.nextInt(stationCount), new VisitMemoryImpl(
//				stationCount), 100000);
		final StationsImpl stations = new StationsImpl(graph, random);
		if (this.trace != null)
			trace.println(stations.toString());
		final ReactiveCommunity community = this.entity.getCommunity();
		final long boardPermissionKey = ReactiveFactoryImpl.REACTIVE_FACTORY
				.newKey();
		final long donePermissionKey = ReactiveFactoryImpl.REACTIVE_FACTORY
				.newKey();
		final LandBoard landBoard = new LandBoardImpl(community,
				boardPermissionKey, donePermissionKey, stations);
		community.setBoard(landBoard, boardPermissionKey);
		final Families families = new Families(stations, random);
		if (this.trace != null)
			trace.println(families.toString());
		final StrategyStat strategyStat = new StrategyStat(families);
		final BehaviorStat behaviorStat = new BehaviorStat();
		final int initialEnergy = BodyAgent.maximumEnergy * 3 / 4;
		final Station initialStation = stations.anyStation(random);
		if (this.trace != null)
			trace.println("Starting from " + initialStation.toString());
		final int familyCount = families.getFamilyCount();
		final AgentFamily[] agentFamilies = new AgentFamily[familyCount];
		for (int k = 0; k < familyCount; k++) {
			String agentId = "a" + (1 + k);
			AgentFamily agentFamily = families.getAgentFamily(k);
			long agentFamilyKey = families.getAgentFamilyKey(k);
			agentFamilies[k] = agentFamily;
			NavigationStrategy navigationStrategy = families.getNavigationStrategy(k);
			GenePool genePool = new GenePool(navigationStrategy);
			BodyAgent bodyAgent = new BodyAgent(agentId, agentFamily,
					agentFamilyKey, genePool, initialEnergy, initialStation,
					strategyStat, behaviorStat, seed, roundTrace);
			community.launchAgent(bodyAgent);
		}
		this.entity.stop();
		int remainingFamilyCount = familyCount;
		for (int step = 0; step < stepCount; step++) {
			this.entity.stop();
			for (int k = 0; k < familyCount; k++) {
				AgentFamily agentFamily = agentFamilies[k];
				if (agentFamily != null
						&& agentFamily.getMemberList().isEmpty()) {
					if (trace != null)
						trace.println("Step " + step + " : "
								+ agentFamily.toString() + " terminates");
					agentFamilies[k] = null;
					remainingFamilyCount--;
				}
			}
			if (remainingFamilyCount <= 1) {
				stepCount = step;
				break;
			}
		}
		this.entity.generateAction(new LandDoneAction(landBoard,
				donePermissionKey));
		do {
			this.entity.stop();
		} while (this.entity.getCommunity().getConnectedEntityCount() > 1);
		this.entity.stop();
		this.entity.stop();
		if (this.trace != null) {
			trace.println(strategyStat.toString());
			trace.println(behaviorStat.toString());
			if (remainingFamilyCount == 1) {
				for (AgentFamily agentFamily : agentFamilies)
					if (agentFamily != null)
						trace.println("Step " + stepCount + " : " + agentFamily
								+ " wins.");
			}
			else
				trace.println("Step " + stepCount + " : "
						+ (remainingFamilyCount == 0 ? "extinction" : "end"));
		}
	}

	///////////////////////
	// Agent implementation
	///////////////////////
	
	public void participate() {
		int seed = new Random().nextInt(1000);
		seed = 666; //TODO to remove or to comment
		if (trace != null)
			trace.println("Seed=" + seed);
		int stepCount = 2000;
		/*
		 * First round.
		 */
		if (trace != null)
			trace.println("--- Round 1");
		final List<List<String>> roundTrace1 = new ArrayList<List<String>>();
		this.executeRound(seed, stepCount, roundTrace1);
		for (List<String> instantTrace : roundTrace1)
			Collections.sort(instantTrace);
		/*
		 * Change community.
		 */
		this.entity.disconnect();
		ReactiveCommunity community = this.launchingContext
				.getReactiveMachine().newCommunity();
		this.entity.connect(community, true);
		/*
		 * Second round.
		 */
		if (trace != null)
			trace.println("--- Round 2");
		final List<List<String>> roundTrace2 = new ArrayList<List<String>>();
		this.executeRound(seed, stepCount, roundTrace2);
		for (List<String> instantTrace : roundTrace2)
			Collections.sort(instantTrace);
		/*
		 * Round comparison.
		 */
		int instantCount = roundTrace1.size() < roundTrace2.size()
				? roundTrace1.size()
				: roundTrace2.size();
		int divergenceInstant = -1;
		for (int k = 0; k < instantCount; k++) {
			List<String> instantTrace1 = roundTrace1.get(k);
			List<String> instantTrace2 = roundTrace2.get(k);
			if (!instantTrace1.equals(instantTrace2)) {
				divergenceInstant = k;
				break;
			}
		}
		if (trace != null) {
			if (divergenceInstant >= 0) {
				if (divergenceInstant == 0)
					throw new IllegalStateException();
				trace.println("Divergence from instant "
						+ (divergenceInstant - 1));
				for (String agentStatus : roundTrace1
						.get(divergenceInstant - 1))
					trace.println(agentStatus);
				trace.println("Round 1 at instant " + divergenceInstant);
				for (String agentStatus : roundTrace1.get(divergenceInstant))
					trace.println(agentStatus);
				trace.println("Round 2 at instant " + divergenceInstant);
				for (String agentStatus : roundTrace2.get(divergenceInstant))
					trace.println(agentStatus);
			} else
				trace.println("--- Rounds 1 and 2 are equivalent.");
		}
	}
}
