/*
 * Copyright (c) 2008, 2011 The_xmloperator_project.  All rights reserved.
 * See terms of use on the project's site (http://www.xmloperator.net/reactive/).
 */
package org.xmloperator.reactive.demo.land.navigation;

import org.xmloperator.reactive.PrintStream;
import org.xmloperator.reactive.demo.land.Graph;
import org.xmloperator.reactive.demo.land.VisitMemory;

/**
 * Navigation strategy that consists in going to the station that was visited a
 * maximum time ago.
 */
public final class DiscoveryNavigationStrategy extends NavigationStrategyImpl {
	private static PrintStream trace = null;

	public static void setTrace(PrintStream trace) {
		DiscoveryNavigationStrategy.trace = trace;
	}

	private final Graph graph;
	private final int[] destinationIndexes;
	private int lastVisitInterval = 0;
	private int period = -1;
	
	public DiscoveryNavigationStrategy(Graph graph) {
		if (graph == null)
			throw new IllegalArgumentException();
		this.graph = graph;
		this.destinationIndexes = new int[graph.getMaxPathCount()];
	}

	/**
	 * Tests the strategy a given step count.
	 * 
	 * @param stationIndex
	 *            the starting station index.
	 * @param visitMemory
	 *            a VisitMemory
	 * @param stepCount
	 *            a step count.
	 * @return the destination station index.
	 */
	public final int test(int stationIndex, VisitMemory visitMemory, int stepCount) {
		int maxInterval = 0;
		int maxIntervalIndex = 0;
		int lastInterval = 0;
		int lastIntervalCount = 0;
		int greaterIntervalCount = 0;
		int greaterCountIndex = 0;
		int greaterCountInterval = 0;
		for (int k = 0; k < stepCount; k++) {
			stationIndex = this.nextStationIndex(stationIndex, visitMemory);
			int interval = this.lastVisitInterval;
			if (interval > maxInterval) {
				maxInterval = interval;
				maxIntervalIndex = k;
			}
			if (interval == lastInterval) {
				lastIntervalCount++;
			}
			else {
				if (lastIntervalCount > greaterIntervalCount) {
					greaterIntervalCount = lastIntervalCount;
					greaterCountIndex = k - 1;
					greaterCountInterval = lastInterval;
				}
				lastInterval = interval;
				lastIntervalCount = 1;
			}
		}
		stationIndex = this.searchPeriod(stationIndex, visitMemory);
		if (trace != null) {
			trace.println("Move=" + maxIntervalIndex + " maxInterval="
					+ maxInterval);
			trace.println("Move=" + greaterCountIndex + " interval="
					+ greaterCountInterval + " count=" + greaterIntervalCount);
			trace.println("Period=" + this.period);
		}
		return stationIndex;
	}

	/**
	 * Searches the period of the station sequence.
	 * 
	 * @param currentStationIndex
	 *            the starting station index.
	 * @param visitMemory
	 *            a VisitMemory
	 * @return the destination station index.
	 */
	private final int searchPeriod(int currentStationIndex,
			VisitMemory visitMemory) {
		int stationCount = this.graph.getStationCount();
		int[] stationsIndex = new int[4 * stationCount];
		for (int k = 0; k < stationsIndex.length; k++) {
			currentStationIndex = this.nextStationIndex(currentStationIndex,
					visitMemory);
			stationsIndex[k] = currentStationIndex;
		}
		int stationIndex = stationsIndex[0];
		int period = 0;
		for (int k = 1; k < stationsIndex.length; k++)
			if (stationsIndex[k] == stationIndex) {
				period = k;
				if (period >= stationCount)
					break;
			}
		if (period > 2 * stationCount)
			return currentStationIndex;
		for (int k = 0; k < period; k++)
			if (stationsIndex[period + k] != stationsIndex[k])
				return currentStationIndex;
		this.period = period;
		return currentStationIndex;
	}
	
	////////////////////////////////////
	// NavigationStrategy implementation
	////////////////////////////////////
	
	public final int nextStationIndex(int stationIndex, VisitMemory visitMemory) {
		if (visitMemory == null)
			throw new IllegalArgumentException();
		int pathCount = this.graph.getPathCount(stationIndex);
		int oldestLastVisit = Integer.MAX_VALUE;
		int oldestDestinationIndex = -1;
		synchronized (this) {
			this.graph.getDestinationIndexes(stationIndex,
					this.destinationIndexes);
			for (int k = 0; k < pathCount; k++) {
				int destinationIndex = this.destinationIndexes[k];
				int lastVisit = visitMemory.getLastVisit(destinationIndex);
				if (lastVisit < oldestLastVisit) {
					oldestLastVisit = lastVisit;
					oldestDestinationIndex = destinationIndex;
				}
			}
		}
		visitMemory.visit(oldestDestinationIndex);
		this.lastVisitInterval = visitMemory.getLastVisitInterval();
		return oldestDestinationIndex;
	}

	////////////////////////
	// Object implementation
	////////////////////////
	
	public String toString() {
		return "discovery";
	}
}
