/*
 * Copyright (c) 2008, 2011 The_xmloperator_project.  All rights reserved.
 * See terms of use on the project's site (http://www.xmloperator.net/reactive/).
 */
package org.xmloperator.reactive.demo.land.sphere.impl;

import java.util.ArrayList;
import java.util.Collections;
import java.util.List;
import java.util.concurrent.atomic.AtomicInteger;

import org.xmloperator.reactive.PrintStream;
import org.xmloperator.reactive.demo.land.sphere.Station;
import org.xmloperator.reactive.demo.land.sphere.StationPath;
import org.xmloperator.reactive.demo.land.sphere.Vector2;
import org.xmloperator.reactive.demo.land.sphere.Vector3;

/**
 * Station implementation.
 */
public class StationImpl implements Station {
	private static AtomicInteger stationCount = new AtomicInteger();
	protected static PrintStream trace = null;

	public static void setTrace(PrintStream trace) {
		StationImpl.trace = trace;
	}

	final int id = stationCount.incrementAndGet();
	private Vector3 location = null;
	private Vector3Impl tangentU = null;
	private Vector3Impl tangentV = null;
	private final List<StationPathImpl> stationPathList = new ArrayList<StationPathImpl>();
	private final List<StationPathImpl> unmodifiableStationPathList = Collections
			.unmodifiableList(this.stationPathList);
	private Vector3 repulsionResultant = null;
	
	public StationImpl() {
	}

	/**
	 * Initializes the paths to the other stations.
	 * @param stationList a list of all the stations, including this one.
	 */
	final void initializePathList(List<? extends Station> stationList) {
		if (stationList.size() < 4)
			throw new IllegalArgumentException();
		this.stationPathList.clear();
		for (int k = 0; k < stationList.size(); k++) {
			Station destination = stationList.get(k);
			if (destination != this) {
				StationPathImpl stationPath = new StationPathImpl(this, destination);
				this.stationPathList.add(stationPath);
			}
		}
	}
	
	/**
	 * Set a new location for this Station.
	 * @param location a Point on a sphere of radius 1.
	 */
	final void setLocation(Vector3 location) {
		if (location == null)
			throw new IllegalArgumentException();
		this.location = location;
		double x = location.getX();
		Vector3Impl start = (x > -0.5 && x < 0.5)
				? new Vector3Impl(1.0, 0.0, 0.0)
				: new Vector3Impl(0.0, 1.0, 0.0);
		this.tangentU = Vector3Util.normalizedVectorProduct(start, location);
		this.tangentV = Vector3Util.normalizedVectorProduct(location,
				this.tangentU);
	}
	
	/**
	 * Updates data after a location change.
	 */
	final void update() {
		for (StationPathImpl stationPath : this.stationPathList)
			stationPath.update();
		Collections.sort(this.stationPathList,
				new StationPathLengthComparator());
		this.computeRepulsionResultant();
	}
	
	private final void computeRepulsionResultant() {
		int n = this.stationPathList.size();
		double du = 0.0;
		double dv = 0.0;
		for (StationPathImpl stationPath : this.stationPathList) {
			double repulsion = stationPath.getRepulsion();
			Vector2 tangent = stationPath.getTangent();
			du += repulsion * tangent.getU();
			dv += repulsion * tangent.getV();
		}
		du /= n;
		dv /= n;
		double dx = du * this.tangentU.getX() + dv * this.tangentV.getX();
		double dy = du * this.tangentU.getY() + dv * this.tangentV.getY();
		double dz = du * this.tangentU.getZ() + dv * this.tangentV.getZ();
		this.repulsionResultant = new Vector3Impl(-dx, -dy, -dz);
	}
	
	/////////////////////////
	// Station implementation
	/////////////////////////
	
	public final Vector3 getLocation() {
		return this.location;
	}

	public final Vector3 getTangentU() {
		return this.tangentU;
	}

	public final Vector3 getTangentV() {
		return this.tangentV;
	}

	public final List<? extends StationPath> getStationPathList() {
		return this.unmodifiableStationPathList;
	}

	public final Vector3 getRepulsionResultant() {
		return this.repulsionResultant;
	}
	
	public final double repulsionSum() {
		double repulsionSum = 0.0;
		for (StationPathImpl stationPath : this.stationPathList)
			repulsionSum += stationPath.getRepulsion();
		return repulsionSum;
	}
	
	////////////////////////
	// Object implementation
	////////////////////////
	
	public String toString() {
		StringBuffer sb = new StringBuffer();
		sb.append("Station");
		if (this.id < 10)
			sb.append('0');
		sb.append(this.id);
		sb.append(this.location);
		for (StationPath stationPath : this.stationPathList) {
			sb.append("[");
			sb.append(stationPath);
			sb.append("]");
		}
		return sb.toString();
	}
}
