/*
 * Copyright (c) 2008, 2011 The_xmloperator_project.  All rights reserved.
 * See terms of use on the project's site (http://www.xmloperator.net/reactive/).
 */
package org.xmloperator.reactive.demo.land.sphere.impl;

import java.text.DecimalFormat;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.Collections;
import java.util.List;
import java.util.Random;

import org.xmloperator.reactive.PrintStream;
import org.xmloperator.reactive.demo.land.sphere.Station;
import org.xmloperator.reactive.demo.land.sphere.StationPath;
import org.xmloperator.reactive.demo.land.sphere.Stations;
import org.xmloperator.reactive.demo.land.sphere.Vector3;

/**
 * Stations implementation.
 */
public class StationsImpl implements Stations {
	private static final double targetTension = 0.001;
	private static final int maxStepCount = 200;
	private static final double[] coefs =
		new double[] { 0.5, 0.1, 0.02, -0.01 };
	private static PrintStream trace = null;

	public static void setTrace(PrintStream trace) {
		StationsImpl.trace = trace;
	}

	/**
	 * Creates and return a directed graph using a given count of stations and a
	 * same count of path from each station.
	 * 
	 * @param stationCount
	 *            a count of stations.
	 * @param pathCount
	 *            a path count from any station.
	 * @param random
	 *            a random number generator.
	 * @return the created graph : sequence of station, for each station :
	 *         sequence of destination indexes.
	 */
	public static final int[] createDirectedGraph(int stationCount, int pathCount,
			Random random) {
		StationsImpl stations;
		do {
			stations = new StationsImpl(stationCount, random);
			stations.minimizeTension(targetTension, maxStepCount);
		} while (!stations.isGraphComplete(pathCount));
		return stations.graph(pathCount);
	}

	/**
	 * Creates and return an undirected net using a max path length between
	 * stations and a variable count of path from each station.
	 * 
	 * @param stationCount
	 *            a count of stations.
	 * @param minPathCount
	 *            a minimum path count from any station.
	 * @param maxPathCount
	 *            a max path count from any station.
	 * @param random
	 *            a random number generator.
	 * @return the created net : sequence of station, for each station :
	 *         sequence of maxPathCount destination indexes, -1 for no path.
	 */
	public static final int[] createUndirectedGraph(int stationCount,
			int minPathCount, int maxPathCount, Random random) {
		StationsImpl stations;
		double radius;
		int maxCount;
		do {
			stations = new StationsImpl(stationCount, random);
			stations.minimizeTension(targetTension, maxStepCount);
			radius = stations.maxPathLength(minPathCount);
			maxCount = stations.maxPathCount(radius);
			if (trace != null) {
				trace.println("Radius: "
						+ new DecimalFormat("0.000").format(radius)
						+ " pathCount: " + maxCount);
			}
		} while (maxCount > maxPathCount);
		return stations.graph(radius, maxPathCount);
	}
	
	private final List<StationImpl> stationList = new ArrayList<StationImpl>();
	private final List<StationImpl> unmodifiableStationList = Collections
			.unmodifiableList(this.stationList);

	/**
	 * Constructor.
	 * @param stationCount a count of stations. Must be greater or equals to 4.
	 * @param random a random number generator.
	 */
	StationsImpl(int stationCount, Random random) {
		if (stationCount < 4)
			throw new IllegalArgumentException();
		for (int k = 0; k < stationCount; k++) {
			Vector3 location;
			do {
				location = new Vector3Impl(random);
			} while (shortestDistance(location) < 1.0e-3);
			StationImpl station = new StationImpl();
			station.setLocation(location);
			this.stationList.add(station);
		}
		for (StationImpl station : this.stationList)
			station.initializePathList(this.stationList);
		this.update();
	}

	private final double shortestDistance(Vector3 location) {
		double shortestDistance = Double.MAX_VALUE;
		for (StationImpl station : this.stationList) {
			double distance = Vector3Util.distance(location,
					station.getLocation());
			if (distance < shortestDistance)
				shortestDistance = distance;
		}
		return shortestDistance;
	}
	
	/**
	 * Determines the StationPath list of each Station.
	 */
	final void update() {
		for (StationImpl station : this.stationList)
			station.update();
	}

	/**
	 * Minimizes the global tension of the stations.
	 * @param maxTension a max tension.
	 * @param maxStepCount a max step count.
	 */
	final void minimizeTension(double maxTension, int maxStepCount) {
		int stationCount = this.stationList.size();
		double tension;
		int count = 0;
		int[] c = new int[coefs.length];
		do {
			double tensionSquareSum = this.moveLocations(c);
			tension = Math.sqrt(tensionSquareSum / stationCount);
			count++;
//			if (trace != null)
//				trace.println("Tension: "
//						+ new DecimalFormat("0.000000").format(tension));
		} while (tension > maxTension && count < maxStepCount);
		if (trace != null) {
//			for (StationImpl station : this.stationList)
//				trace.println(station.toString());
			StringBuffer sb = new StringBuffer();
			for (int k = 0; k < coefs.length; k++)
				sb.append(" " + c[k]);
			trace.println("Count: " + count + sb);
			trace.println("Sum: "
					+ new DecimalFormat("0.0").format(this.repulsionSum()));
		}
	}
	
	/**
	 * Moves the station locations for a best distribution.
	 * @return the tension square sum.
	 */
	final double moveLocations(int c[]) {
		Vector3[] locations = this.getLocations();
		double refTensionSquareSum = this.tensionSquareSum();
		for (int k = 0; k < coefs.length; k++) {
			this.moveLocations(coefs[k]);
			double tensionSquareSum = this.tensionSquareSum();
			if (tensionSquareSum < refTensionSquareSum) {
				c[k]++;
				return tensionSquareSum;
			}
			this.setLocations(locations);
		}
		return 0.0;
	}
	
	private final Vector3[] getLocations() {
		int n = this.stationList.size();
		Vector3[] locations = new Vector3[n];
		for (int k = 0; k < n; k++)
			locations[k] = this.stationList.get(k).getLocation();
		return locations;
	}
	
	private final void setLocations(Vector3[] locations) {
		for (int k = 0; k < locations.length; k++)
			this.stationList.get(k).setLocation(locations[k]);
		this.update();
	}
	
	private final double tensionSquareSum() {
		int n = this.stationList.size();
		double tensionSquareSum = 0.0;
		for (int k = 0; k < n; k++) {
			StationImpl station = this.stationList.get(k);
			Vector3 moveVector = station.getRepulsionResultant();
			double x = moveVector.getX();
			double y = moveVector.getY();
			double z = moveVector.getZ();
			tensionSquareSum += x * x + y * y + z * z;
		}
		return tensionSquareSum;
	}
	
	private final void moveLocations(double coef) {
		int n = this.stationList.size();
		Vector3[] newLocations = new Vector3[n];
		for (int k = 0; k < n; k++) {
			StationImpl station = this.stationList.get(k);
			Vector3 moveVector = station.getRepulsionResultant();
			double x = moveVector.getX() * coef;
			double y = moveVector.getY() * coef;
			double z = moveVector.getZ() * coef;
			{
				Vector3 location = station.getLocation();
				x += location.getX();
				y += location.getY();
				z += location.getZ();
			}
			newLocations[k] = Vector3Util.normalizedVector(x, y, z);
		}
		for (int k = 0; k < n; k++) {
			StationImpl station = this.stationList.get(k);
			station.setLocation(newLocations[k]);
		}
		this.update();
	}

	/**
	 * @return twice the sum of all repulsions.
	 */
	final double repulsionSum() {
		double repulsionSum = 0.0;
		for (StationImpl station : this.stationList)
			repulsionSum += station.repulsionSum();
		return repulsionSum;
	}

	/**
	 * @param pathCount
	 *            a count of path, valid for any Station.
	 * @return true if any Station is accessible from any other by a sequence of
	 *         path. Each path belongs to the first ones of each Station.
	 */
	final boolean isGraphComplete(int pathCount) {
		int n = this.stationList.size();
		if (pathCount < 3 || pathCount > n - 1)
			throw new IllegalArgumentException();
		int id0 = this.stationList.get(0).id;
		List<StationImpl> stations = new ArrayList<StationImpl>();
		boolean[] isMet = new boolean[n];
		for (StationImpl origin : this.stationList) {
			Arrays.fill(isMet, false);
			isMet[origin.id - id0] = true;
			int count = 1;
			stations.add(origin);
			while (!stations.isEmpty()) {
				StationImpl station1 = stations.remove(0);
				List<? extends StationPath> stationPathList = station1
						.getStationPathList();
				for (int k = 0; k < pathCount; k++) {
					StationImpl station2 = (StationImpl) stationPathList.get(k)
							.getDestinationStation();
					int index = station2.id - id0;
					if (!isMet[index]) {
						isMet[index] = true;
						count++;
						stations.add(station2);
					}
				}
			}
			if (count != n)
				return false;
		}
		return true;
	}

	/**
	 * @param pathCount
	 *            a count of path, valid for any Station.
	 * @return a graph obtained by considering for each Station the
	 *         pathCount first accessible station.
	 */
	final int[] graph(int pathCount) {
		int n = this.stationList.size();
		if (pathCount < 3 || pathCount > n - 1)
			throw new IllegalArgumentException();
		int id0 = this.stationList.get(0).id;
		int[] graph = new int[n * pathCount];
		int index = 0;
		for (int s = 0; s < n; s++) {
			List<? extends StationPath> stationPathList = this.stationList.get(
					s).getStationPathList();
			for (int k = 0; k < pathCount; k++)
				graph[index++] = ((StationImpl) stationPathList.get(k)
						.getDestinationStation()).id - id0;
		}
		return graph;
	}

	/**
	 * @param pathCount
	 *            a count of the first paths for any station.
	 * @return the max path length amongst the pathCount first paths of all the
	 *         stations.
	 */
	final double maxPathLength(int pathCount) {
		if (pathCount < 3)
			throw new IllegalArgumentException();
		double maxPathLength = 0.0;
		for (StationImpl station : this.stationList) {
			List<? extends StationPath> stationPathList = station.getStationPathList();
			for (int k = 0; k < pathCount; k++) {
				double pathLength = stationPathList.get(k).getPseudoDistance();
				if (pathLength > maxPathLength)
					maxPathLength = pathLength;
			}
		}
		return maxPathLength;
	}

	/**
	 * @param pathLength
	 *            a max path length.
	 * @return the max path count amongst the paths of length less or equals to
	 *         the given length.
	 */
	final int maxPathCount(double pathlength) {
		int maxPathCount = 0;
		for (StationImpl station : this.stationList) {
			List<? extends StationPath> stationPathList = station.getStationPathList();
			int pathCount = stationPathList.size();
			int maxCount = 0;
			for (int k = 0; k < pathCount; k++) {
				double length = stationPathList.get(k).getPseudoDistance();
				if (length > pathlength) {
					maxCount = k;
					break;
				}
			}
			if (maxCount > maxPathCount)
				maxPathCount = maxCount;
		}
		return maxPathCount;
	}

	/**
	 * @param maxPathLength
	 *            a max distance between connected stations.
	 * @param maxPathCount
	 *            a max count of path, valid for any Station.
	 * @return a graph obtained by considering for each Station the
	 *         stations whose distance is less a given distance.
	 */
	final int[] graph(double maxPathLength, int maxPathCount) {
		int n = this.stationList.size();
		if (maxPathCount < 3 || maxPathCount > n - 1)
			throw new IllegalArgumentException();
		int id0 = this.stationList.get(0).id;
		int[] graph = new int[n * maxPathCount];
		int index = 0;
		for (int s = 0; s < n; s++) {
			List<? extends StationPath> stationPathList = this.stationList.get(
					s).getStationPathList();
			for (int k = 0; k < maxPathCount; k++) {
				StationPath stationPath = stationPathList.get(k);
				StationImpl destination = (StationImpl) stationPath.getDestinationStation();
				graph[index++] = (stationPath.getPseudoDistance() <= maxPathLength)
						? destination.id - id0
						: -1;
			}
		}
		return graph;
	}
	
	//////////////////////////
	// Stations implementation
	//////////////////////////
	
	public List<? extends Station> getStationList() {
		return this.unmodifiableStationList;
	}
}
