/*
 * Copyright (c) 2008, 2011 The_xmloperator_project.  All rights reserved.
 * See terms of use on the project's site (http://www.xmloperator.net/reactive/).
 */
package org.xmloperator.reactive.demo.market.action;

import java.math.BigInteger;

import org.xmloperator.reactive.demo.market.Account;
import org.xmloperator.reactive.demo.market.MarketBoard;
import org.xmloperator.reactive.demo.market.Proposal;

/**
 * Action of delivering an order. Order is given to transfer the delivery and,
 * possibly, some reimbursement.
 */
public final class DeliveryAction extends OperatorAction {
	
	public static final DeliveryAction fullDelivery(long operatorKey,
			OrderAction order, Account deliverySourceAccount) {
		return new DeliveryAction(operatorKey, order, order.orderFactor,
				BigInteger.ZERO, deliverySourceAccount, order.deliveryAccount,
				null, null);
	}

	public static final DeliveryAction fullRefund(long operatorKey,
			OrderAction order, Account reimbursementSourceAccount) {
		return new DeliveryAction(operatorKey, order, BigInteger.ZERO,
				order.orderFactor, null, null, reimbursementSourceAccount,
				order.paymentAccount);
	}
	
	public final OrderAction orderAction;
	public final BigInteger deliveryFactor;
	public final BigInteger reimbursementFactor;
	public final Account deliverySourceAccount;
	public final Account deliveryDestinationAccount;
	public final Account reimbursementSourceAccount;
	public final Account reimbursementDestinationAccount;
	
	public DeliveryAction(long operatorKey, OrderAction orderAction,
			BigInteger deliveryFactor, BigInteger reimbursementFactor,
			Account deliverySourceAccount, Account deliveryDestinationAccount,
			Account reimbursementSourceAccount,
			Account reimbursementDestinationAccount) {
		super(operatorKey, orderAction.proposal);
		if (deliveryFactor.compareTo(BigInteger.ZERO) < 0
				|| reimbursementFactor.compareTo(BigInteger.ZERO) < 0)
			throw new IllegalArgumentException();
		Proposal proposal = orderAction.proposal;
		if (deliveryFactor.compareTo(BigInteger.ZERO) > 0
				&& (deliverySourceAccount.getResource() != proposal
						.getProposedResource() || deliveryDestinationAccount
						.getResource() != proposal.getProposedResource()))
			throw new IllegalArgumentException();
		if (reimbursementFactor.compareTo(BigInteger.ZERO) > 0
				&& (reimbursementSourceAccount.getResource() != proposal
						.getPaymentResource() || reimbursementDestinationAccount
						.getResource() != proposal.getPaymentResource()))
			throw new IllegalArgumentException();
		this.orderAction = orderAction;
		this.deliveryFactor = deliveryFactor;
		this.reimbursementFactor = reimbursementFactor;
		this.deliverySourceAccount = deliverySourceAccount;
		this.deliveryDestinationAccount = deliveryDestinationAccount;
		this.reimbursementSourceAccount = reimbursementSourceAccount;
		this.reimbursementDestinationAccount = reimbursementDestinationAccount;
	}
	
	////////////////////////
	// Action implementation
	////////////////////////
	
	public void execute(long permissionKey) {
		super.execute(permissionKey);
		Proposal proposal = this.orderAction.proposal;
		MarketBoard marketBoard = proposal.getMarketBoard();
		if (this.deliveryFactor.compareTo(BigInteger.ZERO) > 0) {
			BigInteger deliveryUnitCount = this.deliveryFactor
					.multiply(BigInteger.valueOf(proposal
							.getProposedResourceAtomicUnitCount()));
			marketBoard.newTransfer(permissionKey,
					this.operatorKey,
					this.deliverySourceAccount, this.deliveryDestinationAccount,
					deliveryUnitCount).execute();
		}
		if (this.reimbursementFactor.compareTo(BigInteger.ZERO) > 0) {
			BigInteger reimbursementUnitCount = this.reimbursementFactor
					.multiply(BigInteger.valueOf(proposal
							.getPaymentResourceAtomicUnitCount()));
			marketBoard.newTransfer(permissionKey,
					this.operatorKey,
					this.reimbursementSourceAccount,
					this.reimbursementDestinationAccount, reimbursementUnitCount).execute();
		}
	}
	
	////////////////////////
	// Object implementation
	////////////////////////
	
	public String toString() {
		return "Delivery(" + this.orderAction + "," + this.deliveryFactor + ","
				+ this.reimbursementFactor + ")";
	}
}
