/*
 * Copyright (c) 2008, 2011 The_xmloperator_project.  All rights reserved.
 * See terms of use on the project's site (http://www.xmloperator.net/reactive/).
 */
package org.xmloperator.reactive.demo.market.agent;

import java.math.BigInteger;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.List;

import org.xmloperator.reactive.Action;
import org.xmloperator.reactive.ActionMessage;
import org.xmloperator.reactive.Event;
import org.xmloperator.reactive.LaunchingContext;
import org.xmloperator.reactive.Message;
import org.xmloperator.reactive.ReactiveCommunity;
import org.xmloperator.reactive.SimpleMessage;
import org.xmloperator.reactive.agent.AgentImpl;
import org.xmloperator.reactive.demo.market.Anticipation;
import org.xmloperator.reactive.demo.market.Ingredient;
import org.xmloperator.reactive.demo.market.InsufficientSoldException;
import org.xmloperator.reactive.demo.market.MarketBoard;
import org.xmloperator.reactive.demo.market.ProjectedAccount;
import org.xmloperator.reactive.demo.market.Proposal;
import org.xmloperator.reactive.demo.market.Resource;
import org.xmloperator.reactive.demo.market.Transformation;
import org.xmloperator.reactive.demo.market.action.DeliveryAction;
import org.xmloperator.reactive.demo.market.action.OrderAction;
import org.xmloperator.reactive.demo.market.action.TransformAction;
import org.xmloperator.reactive.demo.market.impl.AnticipationImpl;
import org.xmloperator.reactive.exception.ReactiveException;

/**
 * A Transformer executes a given Transformation, sells the produced resources
 * and buys the consumed resources.
 */
public final class Transformer extends AgentImpl {
	private final long operatorKey;
	private final Event proposalMessageEvent;
	private final Event exitMessageEvent;
	private final BigInteger maxTransformationFactor;
	private final Transformation transformation;
	private final List<? extends Ingredient> ingredientList;
	private final List<Mission> missionList = new ArrayList<Mission>();
	private boolean mustTerminate = false;
	private double initialValue = 0.0;
	private BigInteger accumulationFactor = BigInteger.ZERO;
	
	public Transformer(long operatorKey, Event proposalMessageEvent,
			Event exitMessageEvent, Transformation transformation,
			BigInteger maxTransformationFactor) {
		super(false);
		if (proposalMessageEvent == null || exitMessageEvent == null
				|| transformation == null
				|| maxTransformationFactor.compareTo(BigInteger.ONE) < 0)
			throw new IllegalArgumentException();
		this.operatorKey = operatorKey;
		this.proposalMessageEvent = proposalMessageEvent;
		this.exitMessageEvent = exitMessageEvent;
		this.transformation = transformation;
		this.maxTransformationFactor = maxTransformationFactor;
		this.ingredientList = this.transformation.getIngredientList();
		this.addParameter(transformation.getTransformationName());
	}

	public final void addMission(ProjectedAccount purchasedResourceAccount,
			ProjectedAccount proposedResourceAccount) {
		Mission mission = new Mission(this.operatorKey,
				purchasedResourceAccount, proposedResourceAccount);
		if (this.missionList.contains(mission))
			throw new IllegalArgumentException();
		this.missionList.add(mission);
	}
	
	private TransformAction transform() {
		BigInteger transformationFactor = this.maxTransformationFactor;
		for (Ingredient ingredient : this.ingredientList)
			if (ingredient.getResourceUnitCount() < 0) {
				BigInteger availableUnitCount = ingredient.getAffectedAccount()
						.getSoldWithNegativeAnticipations(this.operatorKey);
				BigInteger factor = availableUnitCount.divide(BigInteger
						.valueOf(-ingredient.getResourceUnitCount()));
				if (factor.compareTo(transformationFactor) < 0)
					transformationFactor = factor;
			}
		if (transformationFactor.compareTo(BigInteger.ZERO) <= 0)
			return null;
		this.accumulationFactor = this.accumulationFactor
				.add(transformationFactor);
		return new TransformAction((MarketBoard) this.entity.getCommunity()
				.getBoard(), this.transformation, transformationFactor);
	}
	
	private Anticipation[] newTransformationAnticipations(
			TransformAction transformAction) {
		int ingredientCount = this.ingredientList.size();
		Anticipation[] anticipations = new Anticipation[ingredientCount];
		for (int k = 0; k < ingredientCount; k++) {
			Ingredient ingredient = this.ingredientList.get(k);
			Anticipation anticipation = new AnticipationImpl(transformAction,
					ingredient.getResource(), transformAction.factor
							.multiply(BigInteger.valueOf(ingredient
									.getResourceUnitCount())));
			anticipations[k] = anticipation;
		}
		return anticipations;
	}
	
	private void addTransformationAnticipations(Anticipation[] anticipations) {
		int ingredientCount = this.ingredientList.size();
		for (int k = 0; k < ingredientCount; k++) {
			Ingredient ingredient = this.ingredientList.get(k);
			ingredient.getAffectedAccount().addAnticipation(this.operatorKey,
					anticipations[k]);
		}
	}
	
	private void removeTransformationAnticipations(Anticipation[] anticipations) {
		int ingredientCount = this.ingredientList.size();
		for (int k = 0; k < ingredientCount; k++) {
			Ingredient ingredient = this.ingredientList.get(k);
			ingredient.getAffectedAccount().removeAnticipation(this.operatorKey,
					anticipations[k]);
		}
	}
	
	private final void computeQuantitiesToPurchase() {
		int ingredientCount = this.ingredientList.size();
		BigInteger[] quantitiesToPurchase = new BigInteger[ingredientCount];
		boolean isQuantitiesToPurchaseOk = false;
		BigInteger transformationFactor = this.maxTransformationFactor.multiply(BigInteger.valueOf(2));
		do {
			/*
			 * Tentative of transformation.
			 */
			Arrays.fill(quantitiesToPurchase, BigInteger.ZERO);
			boolean isPortionOk = true;
			for (int k = 0; k < ingredientCount; k++) {
				Ingredient consumedIngredient = this.ingredientList.get(k);
				int consumedUnitCount = consumedIngredient.getResourceUnitCount();
				if (consumedUnitCount < 0) {
					BigInteger availableQuantityToConsume = consumedIngredient
							.getAffectedAccount()
							.getSoldWithNegativeAnticipations(this.operatorKey);
					BigInteger quantityToPurchase = transformationFactor
							.multiply(
									BigInteger.valueOf(-consumedUnitCount)).subtract(
									availableQuantityToConsume);
					if (quantityToPurchase.compareTo(BigInteger.ZERO) <= 0)
						/*
						 * No needs for purchasing.
						 */
						continue;
					quantitiesToPurchase[k] = quantityToPurchase;
					Mission purchaseMission = this
							.getMissionThatPurchaseResource(consumedIngredient
									.getAffectedAccount());
					if (purchaseMission == null)
						throw new ReactiveException(this.transformation
								+ " : no mission defined for purchasing "
								+ consumedIngredient.getResource());
					double purchaseRate = purchaseMission.virtualRate();
					long precision = 1000000000;
					BigInteger nominator = BigInteger.valueOf(precision);
					BigInteger denominator = BigInteger.valueOf(Math
							.round(precision * purchaseRate));
					BigInteger paymentQuantity = quantityToPurchase.multiply(
							nominator).divide(denominator);
					BigInteger availableQuantityToPay = purchaseMission.proposedResourceAccount
							.getSoldWithNegativeAnticipations(this.operatorKey);
					if (paymentQuantity.compareTo(availableQuantityToPay) > 0) {
						/*
						 * Not enough payment resource.
						 */
						isPortionOk = false;
						break;
					}
				}
			}
			if (isPortionOk) {
				isQuantitiesToPurchaseOk = true;
				break;
			}
			transformationFactor = transformationFactor.divide(BigInteger.valueOf(2));
		} while (transformationFactor.compareTo(BigInteger.ZERO) > 0);
		if (!isQuantitiesToPurchaseOk)
			Arrays.fill(quantitiesToPurchase, BigInteger.ZERO);
		for (int k = 0; k < ingredientCount; k++) {
			Ingredient ingredient = this.ingredientList.get(k);
			if (ingredient.getResourceUnitCount() < 0) {
				ProjectedAccount account = ingredient.getAffectedAccount();
				Mission mission = this.getMissionThatPurchaseResource(account);
				mission.setResourceUnitCountToPurchase(quantitiesToPurchase[k]);
			}
		}
	}
	
	private final Mission getMissionThatPurchaseResource(
			ProjectedAccount purchasedResourceAccount) {
		for (Mission mission : this.missionList)
			if (mission.purchasedResourceAccount == purchasedResourceAccount)
				return mission;
		return null;
	}
	
	private Mission concernedMission(Proposal proposal) {
		Resource purchasedResource = proposal.getProposedResource();
		Resource paymentResource = proposal.getPaymentResource();
		for (Mission mission : this.missionList)
			if (mission.concerns(purchasedResource, paymentResource))
				return mission;
		return null;
	}
	
	private boolean areAllOrderListEmpty() {
		for (Mission mission : this.missionList)
			if (!mission.isOrderListEmpty())
				return false;
		return true;
	}

//	private void traceStock() {
//		if (this.trace == null)
//			return;
//		StringBuffer sb = new StringBuffer();
//		sb.append(this.reac);
//		sb.append(" ");
//		sb.append(this.transformation.getTransformationName());
//		sb.append("	sold:");
//		int ingredientCount = this.ingredientList.size();
//		for (int k = 0; k < ingredientCount; k++) {
//			sb.append(" ");
//			Ingredient ingredient = this.ingredientList.get(k);
//			sb.append(ingredient.getResource());
//			sb.append("(");
//			sb.append(ingredient.getAffectedAccount().getResourceUnitCount(
//					this.operatorKey));
//			sb.append(")");
//		}
//		this.trace.println(sb);
//	}
	
	///////////////////////
	// Agent implementation
	///////////////////////

	public void init(LaunchingContext launchingContext) {
		super.init(launchingContext);
		this.initialValue = transformation.getAccountsValue(operatorKey);
	}
	
	public void participate() {
		ReactiveCommunity community = this.entity.getCommunity();
		Event actionMessageEvent = community.getActionMessageEvent();
		while (!this.mustTerminate || !this.areAllOrderListEmpty()) {
			ActionMessage transformMessage = null;
			List<ActionMessage> orderMessageList = new ArrayList<ActionMessage>();
			/*
			 * Transformation.
			 */
			Anticipation[] transformationAnticipations = null;
			TransformAction transformAction = this.transform();
			if (transformAction != null) {
				transformMessage = this.entity
						.generateActionMessage(transformAction);
				transformationAnticipations = this
						.newTransformationAnticipations(transformAction);
				this.addTransformationAnticipations(transformationAnticipations);
			}
			/*
			 * Missions.
			 */
			this.computeQuantitiesToPurchase();
			/*
			 * Listen.
			 */
			List<Event> eventList = this.entity.suspend();
			if (!eventList.isEmpty()) {
				if (eventList.contains(this.exitMessageEvent))
					this.mustTerminate = true;
				/*
				 * Proposal reception.
				 */
				if (eventList.contains(this.proposalMessageEvent))
					for (Message message : this.proposalMessageEvent
							.getMessageList()) {
						Proposal proposal = (Proposal) ((SimpleMessage) message)
								.getObject();
						Mission mission = this.concernedMission(proposal);
						if (mission != null) {
							OrderAction order = mission
									.processProposal(proposal);
							if (order != null)
								orderMessageList.add(this.entity
										.generateActionMessage(order));
						}
					}
				/*
				 * Delivery reception.
				 */
				if (eventList.contains(actionMessageEvent))
					for (Message message : actionMessageEvent.getMessageList())
						if (message.isActionMessage()) {
							Action action = ((ActionMessage) message)
									.getAction();
							if (action instanceof DeliveryAction) {
								OrderAction order = ((DeliveryAction) action).orderAction;
								Mission mission = this
										.concernedMission(order.proposal);
								if (mission != null)
									mission.removeOrder(order);
							}
						}
				this.entity.stop();
			}
			/*
			 * Next instant.
			 */
			if (transformationAnticipations != null)
				this.removeTransformationAnticipations(transformationAnticipations);
			/*
			 * Exceptions processing.
			 */
			if (transformMessage != null
					&& transformMessage.getException() != null)
				throw new ReactiveException(transformMessage.getException());
			for (ActionMessage orderMessage : orderMessageList) {
				Exception exception = orderMessage.getException();
				if (exception != null)
					if (exception instanceof InsufficientSoldException) {
						OrderAction order = (OrderAction) orderMessage.getAction();
						Mission mission = this.concernedMission(order.proposal);
						if (mission != null)
							mission.removeOrder(order);
						if (this.trace != null)
							this.trace.println(this.entity
									+ " has no sufficient sold on "
									+ ((InsufficientSoldException) exception)
											.getAccount() + " for " + order);
					}
					else
						throw new ReactiveException(exception);
			}
		}
		this.entity.stop();
		this.entity.stop();
	}

	public void terminate() {
		super.terminate();
		if (this.transformation.hasAnticipations())
			throw new IllegalStateException();
		if (this.trace != null && this.initialValue > 0.01) {
			double finalValue = transformation
					.getAccountsValue(this.operatorKey);
			double relativeProfit = finalValue / this.initialValue - 1.0;
			trace.println(this.entity + " "
					+ this.transformation.getTransformationName() + " profit("
					+ Math.round(this.initialValue) + ">"
					+ Math.round(finalValue) + "): "
					+ Math.round(100.0 * relativeProfit) + "% factor: "
					+ this.accumulationFactor);
//			this.traceStock();
		}
	}
}
