/*
 * Copyright (c) 2008, 2011 The_xmloperator_project.  All rights reserved.
 * See terms of use on the project's site (http://www.xmloperator.net/reactive/).
 */
package org.xmloperator.reactive.demo.market.impl;

import java.math.BigInteger;
import java.util.ArrayList;
import java.util.Collections;
import java.util.List;
import java.util.Random;

import org.xmloperator.reactive.demo.market.Ingredient;
import org.xmloperator.reactive.demo.market.InsufficientSoldException;
import org.xmloperator.reactive.demo.market.Resource;
import org.xmloperator.reactive.demo.market.Transformation;

/**
 * Transformation implementation.
 */
public final class TransformationImpl implements Transformation {
	private final String transformationName;
	private final List<IngredientImpl> ingredientList = new ArrayList<IngredientImpl>();
	private final List<? extends Ingredient> unmodifiableIngredientList = Collections
			.unmodifiableList(this.ingredientList);

	public TransformationImpl(String transformationName,
			IngredientImpl[] ingredients) {
		if (transformationName == null)
			throw new IllegalArgumentException();
		this.transformationName = transformationName;
		for (IngredientImpl ingredient : ingredients) {
			if (ingredient == null || this.ingredientList.contains(ingredient))
				throw new IllegalArgumentException();
			this.ingredientList.add(ingredient);
		}
	}
	
	/**
	 * Executes this Transformation.
	 * 
	 * @throws InsufficientSoldException
	 *             thrown if one of the consumed resource Account has an
	 *             insufficient sold.
	 */
	void execute(BigInteger factor) throws InsufficientSoldException {
		if (factor.compareTo(BigInteger.ONE) < 0)
			throw new IllegalArgumentException();
		int resourceCount = this.ingredientList.size();
		/*
		 * Consumption.
		 */
		{
			int resourceIndex = 0;
			while (resourceIndex < resourceCount) {
				IngredientImpl ingredient = this.ingredientList.get(resourceIndex);
				if (ingredient.getResourceUnitCount() < 0
						&& !ingredient.affectAccount(factor))
					break;
				resourceIndex++;
			}
			if (resourceIndex < resourceCount) {
				/*
				 * Go back.
				 */
				for (int k = 0; k < resourceIndex; k++)
					this.ingredientList.get(k).affectAccount(factor.negate());
				throw new InsufficientSoldException(this.ingredientList
						.get(resourceIndex).getAffectedAccount());
			}
		}
		/*
		 * Production.
		 */
		for (int k = 0; k < resourceCount; k++) {
			IngredientImpl ingredient = this.ingredientList.get(k);
			if (ingredient.getResourceUnitCount() > 0)
				ingredient.affectAccount(factor);
		}
	}

	/**
	 * Adjusts randomly the coefficients of this Transformation.
	 * @param random a random generator.
	 */
	void adjustProcess(Random random) {
		for (IngredientImpl ingredient : this.ingredientList)
			ingredient.adjustResourceUnitCount(random);
	}

	////////////////////////////////
	// Transformation implementation
	////////////////////////////////
	
	public final String getTransformationName() {
		return this.transformationName;
	}
	
	public final List<? extends Ingredient> getIngredientList() {
		return this.unmodifiableIngredientList;
	}

	public final Ingredient getIngredient(Resource resource) {
		int resourceCount = this.ingredientList.size();
		for (int k = 0; k < resourceCount; k++) {
			Ingredient ingredient = this.ingredientList.get(k);
			if (ingredient.getResource() == resource)
				return ingredient;
		}
		return null;
	}
	
	public final double getAccountsValue(long holderKey) {
		double value = 0.0;
		for (Ingredient ingredient : this.ingredientList)
			value += ingredient.getAffectedAccount().getResourceValue(holderKey);
		return value;
	}
	
	public final boolean hasAnticipations() {
		for (Ingredient ingredient : this.ingredientList)
			if (ingredient.getAffectedAccount().hasAnticipations())
				return true;
		return false;
	}
	
	////////////////////////
	// Object implementation
	////////////////////////
	
	public String toString() {
		StringBuffer sb = new StringBuffer();
		sb.append(this.transformationName);
		sb.append("(");
		boolean isStart = true;
		for (Ingredient ingredient : this.ingredientList) {
			if (isStart)
				isStart = false;
			else
				sb.append(",");
			sb.append(ingredient);
		}
		sb.append(")");
		return sb.toString();
	}
}
