/*
 * Copyright (c) 2008 The_xmloperator_project.  All rights reserved.
 * See terms of use on the project's site (http://www.xmloperator.net/reactive/).
 */
package org.xmloperator.reactive.element.util;

import java.io.OutputStream;
import java.io.OutputStreamWriter;
import java.io.PrintWriter;
import java.io.UnsupportedEncodingException;

import org.xmloperator.reactive.exception.ReactiveException;

import org.xmloperator.reactive.element.exception.UndefinedNamespacePrefixException;
import org.xmloperator.reactive.element.ChildrenConstraint;
import org.xmloperator.reactive.element.DataElement;
import org.xmloperator.reactive.element.RootElement;
import org.xmloperator.reactive.element.Element;
import org.xmloperator.reactive.element.Namespace;
import org.xmloperator.reactive.element.ObjectDataElement;
import org.xmloperator.reactive.element.ParentElement;
import org.xmloperator.reactive.element.TranslatableObject;

/**
 * A DocumentWriter writes a RootElement as an XML document.
 */
public final class DocumentWriter {
	
	public static final String MIME_ENCODING = "UTF-8";
	public static final String JAVA_ENCODING = "UTF8";
	
	/**
	 * Prints a RootElement as an XML 1.0 document.
	 * @param outputStream an OutputStream. It is required.
	 * @param rootElement a RootElement.
	 * @param defaultNamespaceURI the default namespace URI.
	 * @throws UndefinedNamespacePrefixException if a tag name is associated
	 *   to a namespaceURI whose associated prefix is undefined.
	 * @throws AlreadyUsedNamespacePrefixException if a prefix is associated
	 *   to two or more different namespace URI.
	 */
	public static void printDocument(OutputStream outputStream,
			RootElement rootElement,
			String defaultNamespaceURI) {
		if (outputStream == null || rootElement == null)
			throw new IllegalArgumentException();
		try {
			PrintWriter printWriter = new PrintWriter(
					new OutputStreamWriter(outputStream, JAVA_ENCODING));
			new DocumentWriter().print(printWriter, MIME_ENCODING,
					rootElement, defaultNamespaceURI);
		}
		catch (UnsupportedEncodingException exception) {
			exception.printStackTrace();
			throw new ReactiveException("UnsupportedEncodingException");
		}
	}
	
	private PrintWriter out = null;
	private String encodingName = null;
	private String defaultNamespaceURI = null;
	private RootElement rootElement = null;
	
	private DocumentWriter() {
	}
		
	private void print(PrintWriter out, String encodingName,
			RootElement rootElement, String defaultNamespaceURI) {
		if (out == null || encodingName == null || rootElement == null)
			throw new IllegalArgumentException();
		this.out = out;
		this.encodingName = encodingName;
		this.defaultNamespaceURI = defaultNamespaceURI;
		this.rootElement = rootElement;
		printHeading();
		printElement(rootElement, true);
		out.close();
	}

	private void printHeading() {
		this.out.print("<?xml version=\"1.0\" encoding=\"");
		this.out.print(encodingName);
		this.out.println("\" standalone=\"yes\"?>");
	}
	
	private String qualifiedName(Element element) {
		String localName = element.getLocalName();
		String namespaceURI = element.getNamespaceURI();
		if (namespaceURI == null || namespaceURI == this.defaultNamespaceURI)
			return localName;
		for (int k = 0; k < this.rootElement.namespaceCount(); k++) {
			Namespace namespace = this.rootElement.getNamespace(k);
			if (namespace.namespaceURI == namespaceURI)
				return namespace.prefix + ":" + localName;
		}
		throw new UndefinedNamespacePrefixException(namespaceURI);
	}
	
	private void printElement(Element element, boolean addsXmlns) {
		String qualifiedName = qualifiedName(element);
		this.out.print('<');
		this.out.print(qualifiedName);
		boolean isEmptyElement = false;
		if (element.isDataElement()) {
			DataElement dataElement = (DataElement)element;
			String dataType = dataElement.getDataType();
			this.printAttribute("datatype", dataType);
			if (dataType == DataElement.TYPE_OBJECT) {
				TranslatableObject object =
					((ObjectDataElement)dataElement).getData();
				if (object != null)
					this.printAttribute("class",
						object.getRebuilderClass().getName());
				else {
					isEmptyElement = true;
					this.out.print('/');
				}
				this.out.print('>');
				if (object != null) {
					ParentElement translationElement = object.translate();
					//this.printElement(translationElement, false);
					this.printChildren(translationElement);
				}
			}
			else {
				String dataValue = dataElement.stringDataValue();
				if (dataValue.length() == 0) {
					isEmptyElement = true;
					this.out.print('/');
				}
				this.out.print('>');
				this.printText(dataValue);
			}
		}
		else if (element.isParentElement()) {
			if (addsXmlns) {
				if (this.defaultNamespaceURI != null)
				this.printAttribute("xmlns", this.defaultNamespaceURI);
				for (int k = 0; k < this.rootElement.namespaceCount();
						k++) {
					Namespace namespace =
						this.rootElement.getNamespace(k);
					this.printAttribute("xmlns:" + namespace.prefix,
							namespace.namespaceURI);
				}
			}
			ParentElement parentElement = (ParentElement)element;
			ChildrenConstraint childrenConstraint =
				parentElement.getChildrenContraint();
			if (childrenConstraint != null)
				this.printAttribute(
					"cc", childrenConstraint.toString());
			if (parentElement.getFirstChild() == null) {
				isEmptyElement = true;
				this.out.print('/');
			}
			this.out.print('>');
			this.printChildren(parentElement);
		}
		else {
			isEmptyElement = true;
			this.out.print('/');
			this.out.print('>');
		}
		if (!isEmptyElement) {
			this.out.print('<');
			this.out.print('/');
			this.out.print(qualifiedName);
			this.out.print('>');
		}
	}
	
	private void printChildren(ParentElement parentElement) {
		Element child = parentElement.getFirstChild();
		while (child != null) {
			this.printElement(child, false);
			child = child.getNextSibling();
		}
	}

	private void printAttribute(String name, String value) {
		this.out.print(' ');
		this.out.print(name);
		this.out.print('=');
		this.out.print('"');
		for (int k = 0; k < value.length(); k++) {
			char ch = value.charAt(k);
			switch (ch) {
				case '&':
					this.out.print("&amp;");
					break;
				case '<':
					this.out.print("&lt;");
					break;
				case '>':
					this.out.print("&gt;");
					break;
				case '"':
					this.out.print("&quot;");
					break;
				case '\r':
					break;
				case '\n':
					this.out.println();
					break;
				default:
					this.out.print(ch);
			}
		}
		this.out.print('"');
	}

	private void printText(String value) {
		for (int k = 0; k < value.length(); k++) {
			char ch = value.charAt(k);
			switch (ch) {
				case '&':
					this.out.print("&amp;");
					break;
				case '<':
					this.out.print("&lt;");
					break;
				case '>':
					this.out.print("&gt;");
					break;
				case '\r':
					break;
				case '\n':
					this.out.println();
					break;
				default:
					this.out.print(ch);
			}
		}
	}
}
