/*
 * Copyright (c) 2008, 2011 The_xmloperator_project.  All rights reserved.
 * See terms of use on the project's site (http://www.xmloperator.net/reactive/).
 */
package org.xmloperator.reactive.impl;

import java.util.List;
import java.util.concurrent.atomic.AtomicInteger;

import org.xmloperator.reactive.exception.DisconnectedEntityException;
import org.xmloperator.reactive.exception.EntityAlreadyConnectedException;
import org.xmloperator.reactive.exception.NotBaseEventException;
import org.xmloperator.reactive.exception.NotMessageEventException;
import org.xmloperator.reactive.exception.NotOwnThreadException;
import org.xmloperator.reactive.exception.ThreadAlreadyAttachedException;

import org.xmloperator.reactive.Action;
import org.xmloperator.reactive.ActionMessage;
import org.xmloperator.reactive.Event;
import org.xmloperator.reactive.Message;
import org.xmloperator.reactive.PrintStream;
import org.xmloperator.reactive.ReactiveCommunity;
import org.xmloperator.reactive.ReactiveEntity;

/**
 * ReactiveEntity implementation.
 */
public final class ReactiveEntityImpl implements ReactiveEntity {
	private static AtomicInteger reactiveEntityCount = new AtomicInteger();
	private static PrintStream trace = null;
	private static boolean isFullTrace = false;
	
	public static void setTrace(PrintStream trace, boolean isFullTrace) {
		ReactiveEntityImpl.trace = trace;
		ReactiveEntityImpl.isFullTrace = isFullTrace;
	}
	
	private static String listToString(List<?> list) {
		StringBuffer sb = new StringBuffer();
		for (int k = 0; k < list.size(); k++) {
			if (k != 0)
				sb.append(" ");
			sb.append(list.get(k)); 
		}
		return sb.toString();
	}

	private final String entityId = ("A" + String
			.valueOf(ReactiveEntityImpl.reactiveEntityCount.incrementAndGet()))
			.intern();
	private volatile Thread thread = null;
	private volatile int status = ReactiveEntity.STATUS_NOT_CONNECTED;
	private volatile ReactiveCommunityImpl community = null;
	private volatile Event exclusiveMessageEvent = null;
	private volatile ReactiveCommunityImpl.EntityGroup exclusiveEntityGroup;
	
	public ReactiveEntityImpl() {
	}
	
	ReactiveCommunityImpl.EntityGroup getExclusiveEntityGroup() {
		return this.exclusiveEntityGroup;
	}
	
	private void verifyCurrentThread() {
		if (Thread.currentThread() != this.thread)
			throw new NotOwnThreadException();
	}

	private void verifyCurrentThreadAndConnected() {
		if (Thread.currentThread() != this.thread)
			throw new NotOwnThreadException();
		if (this.community == null)
			throw new DisconnectedEntityException();
	}
	
	////////////////////////////////
	// ReactiveEntity implementation
	////////////////////////////////
	
	public void setThread(Thread thread) throws ThreadAlreadyAttachedException {
		if (this.thread != null)
			throw new ThreadAlreadyAttachedException();
		this.thread = thread;
	}
	
	public String getEntityId() {
		return this.entityId;
	}
	
	public int getStatus() {
		return this.status;
	}
	
	public void setStatusWaiting() {
		this.status = ReactiveEntity.STATUS_WAITING;
	}
	
	public void setStatusActive() {
		this.status = ReactiveEntity.STATUS_ACTIVE;
	}

	public ReactiveCommunity getCommunity() {
		return this.community;
	}

	public void connect(ReactiveCommunity community,
			boolean withExclusiveEntityGroup) {
		this.verifyCurrentThread();
		if (this.community != null)
			throw new EntityAlreadyConnectedException();
		if (trace != null)
			trace.println(this + " connect");
		ReactiveCommunityImpl communityI = ((ReactiveCommunityImpl)community); 
		communityI.connect(this);
		this.community = communityI;
		if (withExclusiveEntityGroup) {
			ReactiveCommunityImpl.EntityGroup entityGroup = communityI.getEntityGroup(this);
			if (entityGroup != null) {
				this.exclusiveEntityGroup = entityGroup;
				this.exclusiveMessageEvent = entityGroup.getExclusiveMessageEvent();
			}
			else {
				this.exclusiveMessageEvent = new MessageEventImpl(communityI, this);
				this.exclusiveEntityGroup = communityI.new EntityGroup(this,
						this.exclusiveMessageEvent);
				communityI.putEntityGroup(this, this.exclusiveEntityGroup);
			}
		}
		else
			communityI.removeEntityGroup(this);
		this.status = ReactiveEntity.STATUS_ACTIVE;
	}

	public Event getExclusiveMessageEvent() {
		return this.exclusiveMessageEvent;
	}

	public void generateEvent(Event event) {
		if (event.getMessageList() != null)
			throw new NotBaseEventException();
		this.verifyCurrentThreadAndConnected();
		if (trace != null && isFullTrace)
			trace.println(this + " generate " + event);
		this.community.generateEvent(this, event);
	}
	
	public void generateMessage(Event messageEvent, Message message) {
		if (messageEvent.getMessageList() == null)
			throw new NotMessageEventException();
		this.verifyCurrentThreadAndConnected();
		if (message != null)
			message.setSenderEntityId(this.entityId);
		if (trace != null && isFullTrace
				&& (message == null || message.isTraceable()))
			trace.println(this + " generate " + messageEvent + " " + message);
		this.community.generateMessage(this, messageEvent, message);
	}

	public ActionMessage generateActionMessage(Action action) {
		ActionMessage actionMessage = new ActionMessageImpl(action);
		this.generateMessage(this.community.actionMessageEvent,
				actionMessage);
		return actionMessage;
	}
	
	public void generateAction(Action action) {
		this.verifyCurrentThreadAndConnected();
		if (trace != null && isFullTrace && action.isTraceable())
			trace.println(this + " generate " + action);
		((ActionImpl)action).sign(this);
		this.community.generateAction(this, action);
	}
	
	public List<Event> suspend() {
		this.verifyCurrentThreadAndConnected();
		if (trace != null && isFullTrace)
			trace.println(this + " suspend");
		this.status = ReactiveEntity.STATUS_SUSPENDED;
		List<Event> eventList = this.community.suspend(this);
		this.status = ReactiveEntity.STATUS_ACTIVE;
		if (trace != null && isFullTrace && eventList != null)
			trace.println(this + " receive " + listToString(eventList));
		return eventList;
	}

	public void stop() {
		this.verifyCurrentThreadAndConnected();
		if (trace != null && isFullTrace)
			trace.println(this + " stop");
		this.status = ReactiveEntity.STATUS_STOPPED;
		this.community.stop(this);
		this.status = ReactiveEntity.STATUS_ACTIVE;
	}

	public void disconnect() {
		if (this.community == null)
			return;
		this.verifyCurrentThread();
		if (trace != null && isFullTrace)
			trace.println(this + " disconnect");
		this.community.disconnect(this);
		this.status = ReactiveEntity.STATUS_NOT_CONNECTED;
		this.community = null;
		this.exclusiveEntityGroup = null;
		this.exclusiveMessageEvent = null;
	}
	
	////////////////////////
	// Object implementation
	////////////////////////

	public String toString() {
		return this.entityId;
	}
}
